/**
 * @since 1.0.0
 */
import type { LazyArg } from "@effect/data/Function"
import type * as HashMap from "@effect/data/HashMap"
import type * as List from "@effect/data/List"
import type * as Option from "@effect/data/Option"
import type { Pipeable } from "@effect/data/Pipeable"
import type * as Cause from "@effect/io/Cause"
import type { Effect } from "@effect/io/Effect"
import type * as FiberId from "@effect/io/FiberId"
import type * as FiberRefs from "@effect/io/FiberRefs"
import * as fiberRuntime from "@effect/io/internal/fiberRuntime"
import * as circular from "@effect/io/internal/layer/circular"
import * as internal from "@effect/io/internal/logger"
import * as internalCircular from "@effect/io/internal/logger-circular"
import type * as Layer from "@effect/io/Layer"
import type * as LogLevel from "@effect/io/LogLevel"
import type * as LogSpan from "@effect/io/LogSpan"
import type { Scope } from "@effect/io/Scope"

/**
 * @since 1.0.0
 * @category symbols
 */
export const LoggerTypeId: unique symbol = internal.LoggerTypeId

/**
 * @since 1.0.0
 * @category symbols
 */
export type LoggerTypeId = typeof LoggerTypeId

/**
 * @since 1.0.0
 * @category models
 */
export interface Logger<Message, Output> extends Logger.Variance<Message, Output>, Pipeable {
  readonly log: (
    options: {
      readonly fiberId: FiberId.FiberId
      readonly logLevel: LogLevel.LogLevel
      readonly message: Message
      readonly cause: Cause.Cause<unknown>
      readonly context: FiberRefs.FiberRefs
      readonly spans: List.List<LogSpan.LogSpan>
      readonly annotations: HashMap.HashMap<string, AnnotationValue>
      readonly date: Date
    }
  ) => Output
}

/**
 * @since 1.0.0
 */
export declare namespace Logger {
  /**
   * @since 1.0.0
   * @category models
   */
  export interface Variance<Message, Output> {
    readonly [LoggerTypeId]: {
      readonly _Message: (_: Message) => void
      readonly _Output: (_: never) => Output
    }
  }
}

/**
 * @since 1.0.0
 * @category models
 */
export type AnnotationValue = string | number | boolean

/**
 * @category constructors
 * @since 1.0.0
 */
export const make: <Message, Output>(
  log: (
    options: {
      readonly fiberId: FiberId.FiberId
      readonly logLevel: LogLevel.LogLevel
      readonly message: Message
      readonly cause: Cause.Cause<unknown>
      readonly context: FiberRefs.FiberRefs
      readonly spans: List.List<LogSpan.LogSpan>
      readonly annotations: HashMap.HashMap<string, AnnotationValue>
      readonly date: Date
    }
  ) => Output
) => Logger<Message, Output> = internal.makeLogger

/**
 * @since 1.0.0
 * @category context
 */
export const add: <B>(logger: Logger<unknown, B>) => Layer.Layer<never, never, never> = circular.addLogger

/**
 * @since 1.0.0
 * @category context
 */
export const addEffect: <R, E, A>(effect: Effect<R, E, Logger<unknown, A>>) => Layer.Layer<R, E, never> =
  circular.addLoggerEffect

/**
 * @since 1.0.0
 * @category context
 */
export const addScoped: <R, E, A>(
  effect: Effect<R, E, Logger<unknown, A>>
) => Layer.Layer<Exclude<R, Scope>, E, never> = circular.addLoggerScoped

/**
 * @since 1.0.0
 * @category mapping
 */
export const mapInput: {
  <Message, Message2>(
    f: (message: Message2) => Message
  ): <Output>(self: Logger<Message, Output>) => Logger<Message2, Output>
  <Output, Message, Message2>(
    self: Logger<Message, Output>,
    f: (message: Message2) => Message
  ): Logger<Message2, Output>
} = internal.mapInput

/**
 * Returns a version of this logger that only logs messages when the log level
 * satisfies the specified predicate.
 *
 * @since 1.0.0
 * @category filtering
 */
export const filterLogLevel: {
  (
    f: (logLevel: LogLevel.LogLevel) => boolean
  ): <Message, Output>(self: Logger<Message, Output>) => Logger<Message, Option.Option<Output>>
  <Message, Output>(
    self: Logger<Message, Output>,
    f: (logLevel: LogLevel.LogLevel) => boolean
  ): Logger<Message, Option.Option<Output>>
} = internal.filterLogLevel

/**
 * @since 1.0.0
 * @category mapping
 */
export const map: {
  <Output, Output2>(
    f: (output: Output) => Output2
  ): <Message>(self: Logger<Message, Output>) => Logger<Message, Output2>
  <Message, Output, Output2>(
    self: Logger<Message, Output>,
    f: (output: Output) => Output2
  ): Logger<Message, Output2>
} = internal.map

/**
 * A logger that does nothing in response to logging events.
 *
 * @since 1.0.0
 * @category constructors
 */
export const none: Logger<unknown, void> = internal.none

/**
 * @since 1.0.0
 * @category context
 */
export const remove: <A>(logger: Logger<unknown, A>) => Layer.Layer<never, never, never> = circular.removeLogger

/**
 * @since 1.0.0
 * @category context
 */
export const replace: {
  <B>(that: Logger<unknown, B>): <A>(self: Logger<unknown, A>) => Layer.Layer<never, never, never>
  <A, B>(self: Logger<unknown, A>, that: Logger<unknown, B>): Layer.Layer<never, never, never>
} = circular.replaceLogger

/**
 * @since 1.0.0
 * @category context
 */
export const replaceEffect: {
  <R, E, B>(that: Effect<R, E, Logger<unknown, B>>): <A>(self: Logger<unknown, A>) => Layer.Layer<R, E, never>
  <A, R, E, B>(self: Logger<unknown, A>, that: Effect<R, E, Logger<unknown, B>>): Layer.Layer<R, E, never>
} = circular.replaceLoggerEffect

/**
 * @since 1.0.0
 * @category context
 */
export const replaceScoped: {
  <R, E, B>(
    that: Effect<R, E, Logger<unknown, B>>
  ): <A>(self: Logger<unknown, A>) => Layer.Layer<Exclude<R, Scope>, E, never>
  <A, R, E, B>(
    self: Logger<unknown, A>,
    that: Effect<R, E, Logger<unknown, B>>
  ): Layer.Layer<Exclude<R, Scope>, E, never>
} = circular.replaceLoggerScoped

/**
 * @since 1.0.0
 * @category constructors
 */
export const simple: <A, B>(log: (a: A) => B) => Logger<A, B> = internal.simple

/**
 * @since 1.0.0
 * @category constructors
 */
export const succeed: <A>(value: A) => Logger<unknown, A> = internal.succeed

/**
 * @since 1.0.0
 * @category constructors
 */
export const sync: <A>(evaluate: LazyArg<A>) => Logger<unknown, A> = internal.sync

/**
 * @since 1.0.0
 * @category constructors
 */
export const test: {
  <Message>(input: Message): <Output>(self: Logger<Message, Output>) => Output
  <Message, Output>(self: Logger<Message, Output>, input: Message): Output
} = internalCircular.test

/**
 * @since 1.0.0
 * @category context
 */
export const withMinimumLogLevel: {
  (level: LogLevel.LogLevel): <R, E, A>(self: Effect<R, E, A>) => Effect<R, E, A>
  <R, E, A>(self: Effect<R, E, A>, level: LogLevel.LogLevel): Effect<R, E, A>
} = circular.withMinimumLogLevel

/**
 * Combines this logger with the specified logger to produce a new logger that
 * logs to both this logger and that logger.
 *
 * @since 1.0.0
 * @category zipping
 */
export const zip: {
  <Message2, Output2>(
    that: Logger<Message2, Output2>
  ): <Message, Output>(self: Logger<Message, Output>) => Logger<Message & Message2, readonly [Output, Output2]>
  <Message, Output, Message2, Output2>(
    self: Logger<Message, Output>,
    that: Logger<Message2, Output2>
  ): Logger<Message & Message2, readonly [Output, Output2]>
} = internal.zip

/**
 * @since 1.0.0
 * @category zipping
 */
export const zipLeft: {
  <Message2, Output2>(
    that: Logger<Message2, Output2>
  ): <Message, Output>(self: Logger<Message, Output>) => Logger<Message & Message2, Output>
  <Message, Output, Message2, Output2>(
    self: Logger<Message, Output>,
    that: Logger<Message2, Output2>
  ): Logger<Message & Message2, Output>
} = internal.zipLeft

/**
 * @since 1.0.0
 * @category zipping
 */
export const zipRight: {
  <Message2, Output2>(
    that: Logger<Message2, Output2>
  ): <Message, Output>(self: Logger<Message, Output>) => Logger<Message & Message2, Output2>
  <Message, Output, Message2, Output2>(
    self: Logger<Message, Output>,
    that: Logger<Message2, Output2>
  ): Logger<Message & Message2, Output2>
} = internal.zipRight

/**
 * @since 1.0.0
 * @category constructors
 */
export const defaultLogger: Logger<unknown, void> = fiberRuntime.defaultLogger

/**
 * @since 1.0.0
 * @category constructors
 */
export const logfmtLogger: Logger<unknown, string> = internal.logfmtLogger

/**
 * @since 1.0.0
 * @category constructors
 */
export const stringLogger: Logger<unknown, string> = internal.stringLogger

/**
 * @since 1.0.0
 * @category constructors
 */
export const tracerLogger: Logger<unknown, void> = fiberRuntime.tracerLogger

/**
 * @since 1.0.0
 * @category constructors
 */
export const logFmt: Layer.Layer<never, never, never> = replace(fiberRuntime.defaultLogger, fiberRuntime.logFmtLogger)

/**
 * @since 1.0.0
 * @category context
 */
export const minimumLogLevel: (level: LogLevel.LogLevel) => Layer.Layer<never, never, never> = circular.minimumLogLevel
